import os
import time
import datetime
from subprocess import  call
import numpy as np
import pandas as pd
import seaborn as sns
from scipy import stats
import seaborn as sns
import matplotlib.pyplot as plt
import scipy as sc
from matplotlib.colors import ListedColormap
import scipy.sparse as sparse
from scipy.stats import zscore
import copy
import pickle
from tqdm import tqdm
import importlib
import portalocker
import threading

class basic_function(object):
    @classmethod
    def test(cls):
        print("0")
    @classmethod
    def store_variable_from_pikle_file(cls,file_name,variable):
        with open(file_name,'wb') as f:
            pickle.dump(variable,f,protocol=-1)

    @classmethod
    def store_variable_from_pikle_file_with_lock(cls, file_name, variable,timeout=20):
        with portalocker.Lock(file_name, mode='wb',flags=portalocker.LOCK_EX, timeout=timeout) as f:
            pickle.dump(variable, f, protocol=-1)
    @classmethod
    def load_variable_from_pikle_file(cls,file_name):
        with open(file_name,'rb') as f:
            variable=pickle.load(f)
        return variable

    @classmethod
    def load_variable_from_pikle_file_with_lock(cls, file_name,timeout=20):
        with portalocker.Lock(file_name, mode='rb',flags=portalocker.LOCK_EX, timeout=timeout) as f:
            variable = pickle.load(f)
        return variable

    @classmethod
    def recursive_mkdir(cls,dir):
        if not os.path.exists(dir):
            os.makedirs(dir)

    @classmethod
    def write_out_chr_pos_df(cls,sc_chr_pos_df, raw_data_dir, cell_name, file_name):
        cls.recursive_mkdir(os.path.join(raw_data_dir, cell_name))
        chr_pos_file = os.path.join(raw_data_dir, cell_name, file_name)
        sc_chr_pos_df.to_csv(chr_pos_file, sep='\t')

    @classmethod
    def read_chr_pos(cls,raw_data_dir, cell_name, file_name):
        chr_pos_file = os.path.join(raw_data_dir, cell_name, file_name)
        sc_chr_pos_df = pd.read_table(chr_pos_file, index_col=0)
        return sc_chr_pos_df

    @staticmethod
    def write_out_array_matrix(out_file_name,array_matrix):
        np.savetxt(out_file_name,array_matrix,delimiter='\t')
    @staticmethod
    def read_array_matrix(out_file_name):
        array_matrix=np.genfromtxt(out_file_name,delimiter='\t')
        return array_matrix

    @staticmethod
    def read_df(out_file_name):
        read_df = pd.read_csv(out_file_name, delimiter='\t',index_col=0,header=0)
        return read_df

    @staticmethod
    def write_out_df(out_file_name,one_df):
        one_df.to_csv(out_file_name, sep='\t')

    @staticmethod
    def log_using_time(output_file,using_time):
        using_time = "Using time {}".format(using_time)
        a = np.array([using_time])
        np.savetxt(output_file,a,fmt="%s")

    @staticmethod
    def get_array_with_specifice_index(array,row_index,col_index):
        array=array[np.ix_(row_index, col_index)]
        return array

    @staticmethod
    def read_list(out_file_name):
        import csv
        reader_result = []
        with open(out_file_name, newline='') as f:
            reader = csv.reader(f, delimiter='\t')
            for row in reader:
                reader_result.append(row)
        return out_file_name

    @staticmethod
    def write_out_list(out_file_name,out_list):
        import csv
        with open(out_file_name, 'w', newline='') as f:
            write = csv.writer(f, delimiter='\t')
            write.writerows(out_list)
    @staticmethod
    def get_chr_length_info( chr_length_file):
        chr_length_dic = {}
        # chr1	197195432
        # chr2	181748087
        file_handle = open(chr_length_file, 'r')
        for line in file_handle.readlines():
            temp = line.strip('\n').split('\t')
            chr_length_dic[temp[0]] = int(temp[1])
        return chr_length_dic

    @staticmethod
    def convert_array_to_bin_pair_df(chr_name,bin_length,one_cell_chr_array,data=0):
        matrix = np.ones((bin_length, bin_length))
        matrix = np.triu(matrix, k=1)

        from scipy.sparse import coo_matrix
        matrix_sparse = coo_matrix(matrix)

        df = pd.DataFrame({'col': matrix_sparse.col, 'row': matrix_sparse.row, 'data': matrix_sparse.data})

        df['col'] = df['col'].astype(str)
        df['row'] = df['row'].astype(str)
        df['data']=data
        df['chr_bin_bin']=chr_name+'_'+df['row']+'_'+df['col']
        Example_df=df.copy()
        Example_df.index=Example_df['chr_bin_bin']

        one_cell_chr_array=np.triu(one_cell_chr_array, k=1)
        one_cell_chr_array_sparse = coo_matrix(one_cell_chr_array)

        df = pd.DataFrame({'col': one_cell_chr_array_sparse.col, 'row': one_cell_chr_array_sparse.row, 'data': one_cell_chr_array_sparse.data})
        df['col'] = df['col'].astype(str)
        df['row'] = df['row'].astype(str)
        df['chr_bin_bin']=chr_name+'_'+df['row']+'_'+df['col']
        one_cell_chr_df=df.copy()
        one_cell_chr_df.index=one_cell_chr_df['chr_bin_bin']
        Example_df.loc[one_cell_chr_df.index,'data']=one_cell_chr_df['data'].values
        return Example_df

    @staticmethod
    def generate_Example_bin_pair_df(chr_name,bin_length,data=0):
        matrix = np.ones((bin_length, bin_length))
        matrix = np.triu(matrix, k=1)

        from scipy.sparse import coo_matrix
        matrix_sparse = coo_matrix(matrix)

        df = pd.DataFrame({'col': matrix_sparse.col, 'row': matrix_sparse.row, 'data': matrix_sparse.data})
        df['data']=data
        df['col'] = df['col'].astype(str)
        df['row'] = df['row'].astype(str)
        df['chr_bin_bin']=chr_name+'_'+df['row']+'_'+df['col']
        Example_df=df.copy()
        Example_df.index=Example_df['chr_bin_bin']
        return Example_df

    @staticmethod
    def convert_bin_pair_df_to_array(bin_length, Example_df):
        # 将'chr_bin_bin'列分解为'chr_name', 'row', 'col'
        Example_df[['chr_name', 'row', 'col']] = Example_df['chr_bin_bin'].str.split('_', expand=True)

        # 将'row'和'col'列的数据类型转换为整数，以便用于创建coo_matrix
        Example_df['row'] = Example_df['row'].astype(int)
        Example_df['col'] = Example_df['col'].astype(int)

        # 使用'row'，'col'，和'data'列创建coo_matrix
        from scipy.sparse import coo_matrix
        matrix_sparse = coo_matrix((Example_df['data'], (Example_df['row'], Example_df['col'])), shape=(bin_length, bin_length))

        # 将coo_matrix转换为数组
        array = matrix_sparse.toarray()

        return array
    def get_position(DNA_probe_df):
        ##############################################
        ################start position################
        ##############################################
        '''

        :param DNA_probe_df:
        :return: axis=0, chromID
                  axis=1, regionID_chrom start
        '''
        # chromID_set = np.unique(DNA_probe_df['chromID'])
        temp_chromID = DNA_probe_df['chromID'].values
        chromID_set=np.unique(temp_chromID[np.where(~np.isnan(temp_chromID))])

        start_position = []
        end_position = []
        mid_position = []
        mid_position_Mb = []
        chrom_name = []
        gene_ID = []
        chrom_region_name=[]
        for one_chromID in chromID_set:
            temp = DNA_probe_df[DNA_probe_df['chromID'] == one_chromID]
            one_chrom_name = temp['Chrom'].values[0]
            chrom_name.append(one_chrom_name)

            one_gene_ID = temp['geneID'].values
            one_gene_ID = one_gene_ID.tolist()
            gene_ID.append(one_gene_ID)

            one_chromID_start_position = temp['Start'].values
            one_chromID_end_position = temp['End'].values
            one_chromID_mid_position = one_chromID_start_position + (
                    one_chromID_end_position - one_chromID_start_position) / 2
            start_position.append(one_chromID_start_position)
            end_position.append(one_chromID_end_position)
            mid_position.append(one_chromID_mid_position)
            one_chromID_mid_position_Mb = one_chromID_mid_position / 1000000
            mid_position_Mb.append(one_chromID_mid_position_Mb)

            one_chrom_region_name = []
            one_chr_name = one_chrom_name
            for one_start, one_end in zip(one_chromID_start_position, one_chromID_end_position):
                one_region_name = one_chr_name + ':' + str(one_start) + '-' + str(one_end)
                one_chrom_region_name.append(one_region_name)
            chrom_region_name.append(one_chrom_region_name)
        return start_position,end_position,mid_position,mid_position_Mb,chrom_name,gene_ID,chrom_region_name



import os
import time
from subprocess import  call
import numpy as np
import pandas as pd
import seaborn as sns
from scipy import stats
import seaborn as sns
import matplotlib.pyplot as plt
import scipy as sc
from matplotlib.colors import ListedColormap
import scipy.sparse as sparse
import importlib
import copy
import gc
import datetime
from scipy.stats import zscore
import sys
import scipy.stats as st
import math
from scipy import optimize
from tqdm import tqdm
from scipy.spatial.distance import pdist, squareform
import scanpy as sc
import warnings
from scipy.interpolate import griddata

warnings.filterwarnings('ignore')
from function_basic import basic_function as basic

import ImputeHiFI_Algorithm_v5 as ImputeHiFI

import ImputeHiFI_Algorithm_v6_no_HiC as ImputeHiFI_no_HiC

def preprocess_DNA_HiC(data_dir):
    temp_path = os.path.join(data_dir, 'DNA_FISH_HiC.pkl')
    Example_data_dict=basic.load_variable_from_pikle_file(temp_path)

    cell_name_list=Example_data_dict['cell_name_list']
    all_cell_dna_list=Example_data_dict['all_cell_dna_list']
    Linear_impute_all_cell_dna_list=Example_data_dict['Linear_impute_all_cell_dna_list']
    chr_cell_type_center_value_df=Example_data_dict['chr_cell_type_center_value_df']
    FISH_nearest_neighbors_names_dic=Example_data_dict['FISH_nearest_neighbors_names_dic']
    DNA_FISH_adata=Example_data_dict['DNA_FISH_adata']
    HiC_adata=Example_data_dict['HiC_adata']
    HiC_RNA_correlations_df=Example_data_dict['HiC_RNA_correlations_df']
    DNA_FISH_RNA_HiC_dic_name_tuple_dict=Example_data_dict['DNA_FISH_RNA_HiC_dic_name_tuple_dict']
    chr_bin_length_dic=Example_data_dict['chr_bin_length_dic']
    prxoimtiy_score_suffix = 'proximity_score'
    store_path = os.path.join(data_dir, f'first_seen_probe_data_{prxoimtiy_score_suffix}.h5ad')
    DNA_FISH_adata.write(store_path)
    input_data=(cell_name_list,all_cell_dna_list,Linear_impute_all_cell_dna_list,chr_cell_type_center_value_df,FISH_nearest_neighbors_names_dic,DNA_FISH_adata,HiC_adata,HiC_RNA_correlations_df,DNA_FISH_RNA_HiC_dic_name_tuple_dict,chr_bin_length_dic)
    return input_data


def preprocess_DNA_FISH(input_data_file,data_dir):
    # input_data_file = r"41586_2020_3126_MOESM2_ESM_modify_for_ImputeHiFI.xlsx"
    xls = pd.ExcelFile(os.path.join(data_dir, input_data_file))
    DNA_probe_25kb_df = pd.read_excel(xls, '25-kb resolution')
    start_position, end_position, mid_position, mid_position_Mb, chrom_name, gene_ID, chrom_region_name = basic.get_position(
        DNA_probe_25kb_df)
    data_path = os.path.join(DNA_seqFISH_dir, "DNAseqFISH+{}loci-E14-replicate{}.csv")
    resol = '25kb'
    rep1_data = pd.read_csv(data_path.format(resol, 1))
    rep1_data["cell_type"] = 'rep1'
    rep2_data = pd.read_csv(data_path.format(resol, 2))
    rep2_data["cell_type"] = 'rep2'

    data_path = os.path.join(DNA_seqFISH_dir, "DNAseqFISH+{}loci-E14-clonal{}hr.csv")
    clone_24_data = pd.read_csv(data_path.format(resol, 24))
    clone_24_data["cell_type"] = '24hr'
    clone_48_data = pd.read_csv(data_path.format(resol, 48))
    clone_48_data["cell_type"] = '48hr'

    DNA_pos_data = pd.concat([rep1_data, rep2_data, clone_24_data, clone_48_data], ignore_index=True, sort=False)
    DNA_pos_data = DNA_pos_data[(DNA_pos_data["labelID"] == 0) | (DNA_pos_data["labelID"] == 1)]
    DNA_pos_data['cell_name'] = DNA_pos_data['fov'].astype(str) + '_' + DNA_pos_data['channel'].astype(str) + '_' + \
                                DNA_pos_data['cellID'].astype(str) + '_' + DNA_pos_data['cell_type'].astype(str)
    temp_df = DNA_pos_data.copy()
    unique_cells = temp_df.drop_duplicates(subset='cell_name', keep='first')

    unique_cells = unique_cells[['cell_name', 'cell_type']]

    unique_cells.reset_index(inplace=True, drop=True)

    unique_cells['cell_index'] = list(unique_cells.index)
    cell_name_array = unique_cells['cell_name'].values
    DNA_pos_data['regionID'] = DNA_pos_data['regionID (hyb1-60)']
    zxys_df = DNA_pos_data[['chromID', 'labelID', 'regionID', 'cell_type', 'cell_name', 'z', 'x', 'y']]
    template_dfs = {}

    # 为每个chromID生成一个DataFrame
    for chromID in range(1, 21):  # chromID从1到20
        # 从原始DataFrame中选出对应chromID的regionID
        regions = np.arange(1, 61)

        # 创建新的DataFrame
        template_df_0 = pd.DataFrame({
            'chromID': chromID,
            'regionID': regions,
            'labelID': 0,
            'x': np.nan,
            'y': np.nan,
            'z': np.nan
        })
        template_df_0.index = template_df_0['chromID'].astype(str) + '_' + template_df_0['regionID'].astype(str) + '_' + \
                              template_df_0['labelID'].astype(str)
        template_df_1 = pd.DataFrame({
            'chromID': chromID,
            'regionID': regions,
            'labelID': 1,
            'x': np.nan,
            'y': np.nan,
            'z': np.nan
        })
        template_df_1.index = template_df_1['chromID'].astype(str) + '_' + template_df_1['regionID'].astype(str) + '_' + \
                              template_df_1['labelID'].astype(str)
        # 将新的DataFrame存储到字典中
        template_dfs[chromID] = [template_df_0, template_df_1]
    zxys_df['chromID_regionID_labelID'] = zxys_df['chromID'].astype(str) + '_' + zxys_df['regionID'].astype(str) + '_' + \
                                          zxys_df['labelID'].astype(str)
    zxys_df['chromID'] = zxys_df['chromID'].astype(int)
    zxys_df['labelID'] = zxys_df['labelID'].astype(int)
    all_cell_dna_list = []
    for one_cell_name in tqdm(cell_name_array):
        temp_one_cell_df = zxys_df[zxys_df['cell_name'] == one_cell_name]
        temp_one_cell_df.index = temp_one_cell_df['chromID_regionID_labelID']
        one_cell_list = []
        one_cell_dna_dfs = copy.deepcopy(template_dfs)
        for chromID in range(1, 21):
            one_chrom_list = []
            for labelID in [0, 1]:
                one_chromID_temp_df = temp_one_cell_df[
                    (temp_one_cell_df['chromID'] == chromID) & (temp_one_cell_df['labelID'] == labelID)]
                one_cell_dna_dfs[chromID][labelID].loc[one_chromID_temp_df.index, ['x', 'y', 'z']] = \
                one_chromID_temp_df[['x', 'y', 'z']]
                one_chrom_list.append(one_cell_dna_dfs[chromID][labelID][['z', 'x', 'y']].values)
            one_cell_list.append(one_chrom_list)
        all_cell_dna_list.append(one_cell_list)
    cluster_label = list(unique_cells['cell_type'].values)
    cluster_label_name_dic = {one_label: one_label for one_label in np.unique(cluster_label)}
    chrom_ids = list(unique_cells['cell_name'].values)
    data_dna_zxy = all_cell_dna_list
    keys_list = ['chrom_ids',
                 'region_names',
                 'start_position',
                 'end_position',
                 'mid_position',
                 'mid_position_Mb',
                 'dna_zxys',
                 'chrom_name',
                 'gene_ID',
                 'cluster_label',
                 'cluster_label_name'
                 ]
    values_list = [chrom_ids,
                   chrom_region_name,
                   start_position,
                   end_position,
                   mid_position,
                   mid_position_Mb,
                   data_dna_zxy,
                   chrom_name,
                   gene_ID,
                   cluster_label,
                   cluster_label_name_dic
                   ]
    data = {}
    for one_key, one_value in zip(keys_list, values_list):
        data[one_key] = one_value
    data_pkl_path = 'data.pkl'
    # data_dir=r"/mnt/disk1/scfan/data/Cai_21_Nature/DNA_25kb"
    basic.store_variable_from_pikle_file(os.path.join(data_dir, data_pkl_path),
                                         data)
def generate_Impute_result(chr_bin_length_dic, base_coef,short_range_HiC_weight_bound,
                      long_range_HiC_weight_bound,all_range_diag_sum_weight_HiC_bound,
                        short_range_diag,background_cell_num,
                               HiC_RNA_correlations_df, DNA_FISH_RNA_HiC_dic_name_tuple,
                               DNA_FISH_name_list, missing_rate_df, DNA_FISH_adata,
                               FISH_nearest_neighbors_names_dic,use_mean_count):
    Impute_DNA_FISH_df_dict = {}
    for one_FISH_name in tqdm(DNA_FISH_name_list):
        one_RNA_name, multi_HiC_name_list = DNA_FISH_RNA_HiC_dic_name_tuple[one_FISH_name]
        one_cell_DNA_RNA_confidence = 1
        multi_HiC_FISHconfid_array = np.array([one_cell_DNA_RNA_confidence] * len(multi_HiC_name_list))

        multi_HiC_RNAcoef_array = HiC_RNA_correlations_df.loc[multi_HiC_name_list, one_RNA_name].values
        multi_HiC_RNAcoef_array[multi_HiC_RNAcoef_array <= 0] = base_coef
        weight_multi_HiC_RNAcoef_array = multi_HiC_RNAcoef_array / np.sum(multi_HiC_RNAcoef_array)
        #     print(multi_HiC_RNAcoef_array,weight_multi_HiC_RNAcoef_array)
        #     break

        # chr_name='chr1'
        Impute_df_dict = {}
        for chr_name in chr_bin_length_dic.keys():
            DNA_FISH_df, diag_sum_prop, DNA_FISH_all_count, Common_Example_df, diag_array, bin_length = ImputeHiFI.generate_DNA_FISH_df(
                one_FISH_name, chr_name, missing_rate_df, DNA_FISH_adata)

            HiC_df, HiC_diag_sum_prop_dict = ImputeHiFI.generate_HiC_df(diag_array, Common_Example_df,
                                                                        multi_HiC_name_list, chr_name, bin_length,
                                                                        weight_multi_HiC_RNAcoef_array, HiC_adata)

            bg_df, bg_diag_sum_prop_dict, bg_all_count = ImputeHiFI.generate_background_df(diag_array,
                                                                                           Common_Example_df,
                                                                                           one_FISH_name, chr_name,
                                                                                           bin_length, DNA_FISH_adata,
                                                                                           FISH_nearest_neighbors_names_dic,
                                                                                           background_cell_num)

            DNA_Impute_FISH_df, diag_learning_weight_dict, diag_learning_weight_for_diag_sum_dict, diag_init_loss, diag_learning_loss = ImputeHiFI.optimize_and_impute_data(
                diag_array, Common_Example_df, DNA_FISH_df, HiC_df, bg_df,
                initial_guess, short_range_diag, short_range_HiC_weight_bound,
                long_range_HiC_weight_bound, all_range_diag_sum_weight_HiC_bound,
                HiC_diag_sum_prop_dict, bg_diag_sum_prop_dict, bg_all_count,use_mean_count)
            Impute_df_dict[chr_name] = DNA_Impute_FISH_df

        #         one_array=ImputeHiFI.convert_bin_pair_df_to_array(bin_length,DNA_FISH_df)
        #         cmap = plt.cm.seismic
        #         cmap.set_bad('#8a898a')
        #         sns.heatmap(one_array,cmap=cmap)
        #         plt.title(chr_name)
        #         plt.show()

        #         one_array=ImputeHiFI.convert_bin_pair_df_to_array(bin_length,DNA_Impute_FISH_df)
        #         cmap = plt.cm.seismic
        #         cmap.set_bad('#8a898a')
        #         sns.heatmap(one_array,cmap=cmap,vmax=1)
        #         plt.title(chr_name)
        #         plt.show()
        #         break
        one_cell_Impute_df = pd.concat(Impute_df_dict.values(), axis=0)
        Impute_DNA_FISH_df_dict[one_FISH_name] = one_cell_Impute_df
    return Impute_DNA_FISH_df_dict

def get_infer3D_result(chr_bin_length_dic, chr_name,chr_name_list,all_FISH_name,
                          ImputeHiFI_adata,ImputeHiFI_diff_genes_adata,all_cell_dna_list,Linear_impute_all_cell_dna_list,
                          chr_cell_type_center_value_df,chrom_ids_to_index_dict,loci_pair_type,
                          ):
    optimize_all_cell_dna_list=copy.deepcopy(Linear_impute_all_cell_dna_list)
    for one_FISH_name in tqdm(all_FISH_name):
        one_cell_type = ImputeHiFI_adata.obs.loc[one_FISH_name]['cell_type']
        sigma = chr_cell_type_center_value_df.loc[chr_name + one_cell_type]['Center_Value']

        proximity_array = ImputeHiFI.get_proximity_array(ImputeHiFI_adata, one_FISH_name, chr_name, chr_bin_length_dic,
                                                         set_min_limit=False, min_limit=0.001, set_max_limit=True,
                                                         max_limit=1.0)
        linear_zxys = ImputeHiFI.infer_3D_position.get_specific_dan_zxys(chrom_ids_to_index_dict, one_FISH_name, chr_name_list, chr_name,
                                            Linear_impute_all_cell_dna_list)
        not_imputed_zxys = ImputeHiFI.infer_3D_position.get_specific_dan_zxys(chrom_ids_to_index_dict, one_FISH_name, chr_name_list, chr_name,
                                                 all_cell_dna_list)

        K_imputed = proximity_array.copy()
        initial_guess = linear_zxys.copy()

        nan_position_array=ImputeHiFI.infer_3D_position.get_nan_positions(chrom_ids_to_index_dict,
                                                                          one_FISH_name, chr_name_list,
                                                                          chr_name, all_cell_dna_list)

        not_imputed_all_x = not_imputed_zxys.copy()
        need_imputed_x_initial_guess = initial_guess[nan_position_array].copy()
        # bounds = ImputeHiFI.infer_3D_position.get_bounds(not_nan_array, initial_guess)
        bounds = None
        # optimze_zxys = ImputeHiFI.infer_3D_position_all.optimize_coordinates(K_imputed, sigma, initial_guess,
        #                                                                      maxfun=5000, maxiter=5000, bounds=bounds,
        #                                                                      bias=None, verbose=False)
        optimze_zxys=ImputeHiFI.infer_3D_position.optimize_coordinates(K_imputed, sigma, need_imputed_x_initial_guess,
                                                          not_imputed_all_x, nan_position_array, maxfun=5000,
                                                          maxiter=5000, bounds=bounds, bias=None, verbose=False)
        if loci_pair_type != 'no_diff':
            # print("Fine tune with leiden find import loci pair")
            #fine tune with leiden find import loci pair
            proximity_array = ImputeHiFI.get_proximity_array(ImputeHiFI_diff_genes_adata, one_FISH_name, chr_name, chr_bin_length_dic,
                                                             set_min_limit=False, min_limit=0.001, set_max_limit=True,
                                                             max_limit=1.0)
            K_imputed = proximity_array.copy()
            initial_guess = optimze_zxys.copy()
            not_imputed_all_x = not_imputed_zxys.copy()
            need_imputed_x_initial_guess = initial_guess[nan_position_array].copy()
            optimze_zxys=ImputeHiFI.infer_3D_position.optimize_coordinates(K_imputed, sigma, need_imputed_x_initial_guess,
                                                              not_imputed_all_x, nan_position_array, maxfun=5000,
                                                              maxiter=5000, bounds=bounds, bias=None, verbose=False)


        cell_idx = chrom_ids_to_index_dict[one_FISH_name]
        chr_idx = chr_name_list.index(chr_name)
        optimize_all_cell_dna_list[cell_idx][chr_idx] = optimze_zxys

    corr_dict = {}
    corr_dict['optimize_all_cell_dna_list']=optimize_all_cell_dna_list
    one_chr_corr_dict={}
    one_chr_corr_dict[chr_name]=corr_dict
    return chr_name,one_chr_corr_dict
def run_ImputeHiFI_mode_1(data_dir,input_data,HiC_weight_bound=0.4,neighbor_cell_num=100):
    (cell_name_list, all_cell_dna_list, Linear_impute_all_cell_dna_list, chr_cell_type_center_value_df,
     FISH_nearest_neighbors_names_dic, DNA_FISH_adata, HiC_adata, HiC_RNA_correlations_df,
     DNA_FISH_RNA_HiC_dic_name_tuple_dict, chr_bin_length_dic)=input_data
    kernerl_number = 32
    DNA_FISH_name_list=cell_name_list
    DNA_FISH_name_split_list = np.array_split(DNA_FISH_name_list, kernerl_number)
    chr_cell_type_center_value_df = chr_cell_type_center_value_df.set_index('Chr_Cell_Type')
    import multiprocessing as mp
    num_threads = kernerl_number
    base_coef = 0.001
    HiC_weight = 0.4
    FISH_weight = 0.6
    short_range_HiC_weight_bound = HiC_weight_bound
    long_range_HiC_weight_bound = 0.1
    all_range_diag_sum_weight_HiC_bound = 0.1
    short_range_diag = 20
    initial_guess = [HiC_weight, FISH_weight]
    verbose = False
    background_cell_num = neighbor_cell_num
    para=('bg_n'+str(background_cell_num)+'_sdn'+str(short_range_diag)+'_sd'+str(all_range_diag_sum_weight_HiC_bound)+
          '_sr'+str(short_range_HiC_weight_bound)+'_lr'+str(long_range_HiC_weight_bound))
    parameter_dic={}
    parameter_dic['base_coef']=base_coef
    parameter_dic['HiC_weight']=HiC_weight
    parameter_dic['FISH_weight'] = FISH_weight
    parameter_dic['short_range_HiC_weight_bound'] = short_range_HiC_weight_bound
    parameter_dic['long_range_HiC_weight_bound'] = long_range_HiC_weight_bound
    parameter_dic['all_range_diag_sum_weight_HiC_bound'] = all_range_diag_sum_weight_HiC_bound
    parameter_dic['short_range_diag'] = short_range_diag
    parameter_dic['background_cell_num']=background_cell_num

    _domain_args = [( chr_bin_length_dic, base_coef,short_range_HiC_weight_bound,
                      long_range_HiC_weight_bound,all_range_diag_sum_weight_HiC_bound,
                        short_range_diag,background_cell_num,
                               HiC_RNA_correlations_df, DNA_FISH_RNA_HiC_dic_name_tuple,
                               DNA_FISH_name_list, missing_rate_df, DNA_FISH_adata,
                               FISH_nearest_neighbors_names_dic,use_mean_count)
                    for DNA_FISH_name_list in DNA_FISH_name_split_list]
    with mp.Pool(num_threads) as domain_pool:
        domain_results = domain_pool.starmap(generate_Impute_result, _domain_args)
        domain_pool.close()
        domain_pool.join()
        domain_pool.terminate()

    result_dic = {}
    for result in domain_results:
        print("Collect result")
        for cell_name,df in result.items():
            result_dic[cell_name]=df
    Example_dict = {}
    for chr_name, bin_length in chr_bin_length_dic.items():
        Example_dict[chr_name] = ImputeHiFI.generate_Example_bin_pair_df(chr_name, bin_length)

    all_index_df = pd.concat(Example_dict.values(), axis=0)
    all_cell_df=pd.DataFrame(index=all_index_df.index)
    for cell_name,df in result_dic.items():
        all_cell_df[cell_name]=df.loc[all_cell_df.index,'data'].values
    ImputeHiFI_adata=sc.AnnData(all_cell_df.T)
    ImputeHiFI_adata.obs = DNA_FISH_adata.obs.loc[ImputeHiFI_adata.obs_names]
    ImputeHiFI_adata.uns['parameter'] = parameter_dic
    n_loci_pair=6000
    loci_pair_type='leiden_diff'
    if loci_pair_type == 'leiden_diff':
        diff_genes,not_diff_genes,non_diff_genes_mask=ImputeHiFI.get_diffgene_with_leiden(ImputeHiFI_adata.copy(), n_loci_pair=n_loci_pair)

    if loci_pair_type == 'highly_variable_loci_pair':
        diff_genes, not_diff_genes, non_diff_genes_mask = ImputeHiFI.get_highly_variable_genes(ImputeHiFI_adata.copy(),
                                                                                          n_loci_pair=n_loci_pair)
    if loci_pair_type == 'no_diff':
        diff_genes = []
        not_diff_genes = []
        non_diff_genes_mask = np.zeros(ImputeHiFI_adata.shape[1], dtype=bool)
    # sc.pp.normalize_total(ImputeHiFI_adata, target_sum=1e4)
    ImputeHiFI_diff_genes_adata=ImputeHiFI_adata.copy()
    ImputeHiFI_diff_genes_adata.X[:, non_diff_genes_mask] =0

    chr_name_list=list(chr_bin_length_dic.keys())
    kernerl_number=1
    chr_name_split_list = np.array_split(chr_name_list, kernerl_number)

    import multiprocessing as mp
    num_threads = kernerl_number

    _domain_args = [( chr_bin_length_dic,chr_name_part[0],chr_name_list,all_FISH_name,
                          ImputeHiFI_adata,ImputeHiFI_diff_genes_adata,all_cell_dna_list,
                      Linear_impute_all_cell_dna_list,
                          chr_cell_type_center_value_df,chrom_ids_to_index_dict,loci_pair_type)
                    for chr_name_part in chr_name_split_list]
    with mp.Pool(num_threads) as domain_pool:
        domain_results = domain_pool.starmap(get_infer3D_result, _domain_args)
        domain_pool.close()
        domain_pool.join()
        domain_pool.terminate()


    finall_optimize_all_cell_dna_list=copy.deepcopy(all_cell_dna_list)
    for result in domain_results:
        print("Collect result")
        chr_name,one_chr_corr_dict=result
        optimize_all_cell_dna_list=one_chr_corr_dict[chr_name]['optimize_all_cell_dna_list']
        chr_idx = chr_name_list.index(chr_name)
        for cell_idx in range(len(optimize_all_cell_dna_list)):
            finall_optimize_all_cell_dna_list[cell_idx][chr_idx]=optimize_all_cell_dna_list[cell_idx][chr_idx]
    final_ImputeHiFI_data=copy.deepcopy(data)
    final_ImputeHiFI_data['all_cell_dna_list']=finall_optimize_all_cell_dna_list
    final_ImputeHiFI_data['ledien_diff_gene'] = diff_genes
    temp_path = os.path.join(data_dir, "Impute_data.pkl")

    basic.store_variable_from_pikle_file(temp_path, final_ImputeHiFI_data)


    data = final_ImputeHiFI_data

    kernerl_number = 1
    chr_name_split_list = np.array_split(chr_name_list, kernerl_number)

    import multiprocessing as mp
    num_threads = kernerl_number

    _domain_args = [(data, chr_name_part[0], chr_bin_length_dic)
                    for chr_name_part in chr_name_split_list]
    with mp.Pool(num_threads) as domain_pool:
        domain_results = domain_pool.starmap(generate_spatial_bin_pair_df_one_chr, _domain_args)
        domain_pool.close()
        domain_pool.join()
        domain_pool.terminate()
    all_cell_df_dic = {}
    for result in domain_results:
        chr_name, temp_all_cell_df_dic = result
        all_cell_df_dic[chr_name] = temp_all_cell_df_dic[chr_name]
    all_cell_df = pd.concat(all_cell_df_dic.values())

    adata = sc.AnnData(all_cell_df.T)

    temp_path = os.path.join(data_dir,"Impute_data_spatial_dist.h5ad")
    adata.write(temp_path)

    DNA_FISH_spad_adata = adata


    kernerl_number = 1
    chr_name_split_list = np.array_split(chr_name_list, kernerl_number)
    # store_dir = r"/mnt/disk1/scfan/data/Cai_21_bioRxiv/ImputHiFI/first_seen_probe_data"
    import multiprocessing as mp
    num_threads = kernerl_number

    _domain_args = [
        (DNA_FISH_spad_adata, chr_name_part[0], chr_bin_length_dic, chr_cell_type_center_value_df)
        for chr_name_part in chr_name_split_list]
    with mp.Pool(num_threads) as domain_pool:
        domain_results = domain_pool.starmap(generate_proximity_bin_pair_df_one_chr, _domain_args)
        domain_pool.close()
        domain_pool.join()
        domain_pool.terminate()
    all_cell_df_dic = {}
    for result in domain_results:
        chr_name, temp_all_cell_df_dic = result
        all_cell_df_dic[chr_name] = temp_all_cell_df_dic[chr_name]
    all_cell_df = pd.concat(all_cell_df_dic.values())
    adata = sc.AnnData(all_cell_df.T)
    temp_path = os.path.join(data_dir,"Impute_data_proximity_score.h5ad")
    adata.write(temp_path)

def run_ImputeHiFI_mode_2(first_seen_probe_data_dir,neighbor_cell_num=100):
    # current_probe_data_dir = r"/mnt/disk1/scfan/data/Cai_21_Nature/DNA_25kb"
    # first_seen_probe_data_dir = r"/mnt/disk1/scfan/data/Cai_21_Nature/DNA_25kb/Impute/"
    basic.recursive_mkdir(first_seen_probe_data_dir)

    ImputeHiFI_data_dir = os.path.join(first_seen_probe_data_dir, 'ImputeHiFI')
    basic.recursive_mkdir(ImputeHiFI_data_dir)

    ### proxi adata

    FISH_RMSD_name = 'All_cell_RNA_RMSD_dict'
    version = 'no_HiC_RNA_FISH_RMSD_not_use_mean_count'
    first_seen_probe_data_path = os.path.join(first_seen_probe_data_dir, 'first_seen_probe_data.pkl')
    first_seen_probe_data = basic.load_variable_from_pikle_file(first_seen_probe_data_path)

    chr_cell_type_center_value_df = first_seen_probe_data['chr_cell_type_center_value_df']
    chr_name_list = first_seen_probe_data['chr_name_list']
    cell_name_list=first_seen_probe_data['cell_name_list']
    cell_number = first_seen_probe_data['cell_number']
    all_cell_dna_list = first_seen_probe_data['all_cell_dna_list']
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']
    chrom_index_to_ids_dict = first_seen_probe_data['chrom_index_to_ids_dict']
    adata_obs_df = first_seen_probe_data['adata_obs_df']
    spaital_dist_suffix = first_seen_probe_data['spaital_dist_suffix']
    prxoimtiy_score_suffix = first_seen_probe_data['prxoimtiy_score_suffix']
    missing_rate_df = first_seen_probe_data['missing_rate_df']
    FISH_nearest_neighbors_dic = first_seen_probe_data[FISH_RMSD_name]

    store_path = os.path.join(first_seen_probe_data_dir, f'first_seen_probe_data_{prxoimtiy_score_suffix}.h5ad')
    DNA_FISH_proxi_adata = sc.read(store_path)

    importlib.reload(ImputeHiFI_no_HiC)
    base_coef = 0.001
    HiC_weight = 0
    FISH_weight = 1
    short_range_HiC_weight_bound = 0.0
    long_range_HiC_weight_bound = 0.0
    all_range_diag_sum_weight_HiC_bound = 0.0
    short_range_diag = 0
    initial_guess = [HiC_weight, FISH_weight]
    verbose = False
    background_cell_num = neighbor_cell_num
    use_mean_count = 0
    para = ('bg_n' + str(background_cell_num) + '_sdn' + str(short_range_diag) + '_sd' + str(
        all_range_diag_sum_weight_HiC_bound) +
            '_sr' + str(short_range_HiC_weight_bound) + '_lr' + str(long_range_HiC_weight_bound))

    store_path = os.path.join(ImputeHiFI_data_dir, str(version) + "_proximity_score_adata_" + str(para) + ".h5ad")

    ImputeHiFI_no_HiC.ImputeHiFi_mode2_mode3.get_Impute_proximity_score(DNA_FISH_proxi_adata, missing_rate_df,
                                                                        cell_name_list, chr_bin_length_dic, base_coef,
                                                                        HiC_weight, FISH_weight, initial_guess,
                                                                        short_range_HiC_weight_bound,
                                                                        long_range_HiC_weight_bound,
                                                                        all_range_diag_sum_weight_HiC_bound,
                                                                        short_range_diag, background_cell_num,
                                                                        use_mean_count, FISH_nearest_neighbors_dic,
                                                                        store_path, kernerl_number=20)


    ### Impute 3D data

    first_seen_probe_data.keys()

    importlib.reload(ImputeHiFI)
    store_path = os.path.join(ImputeHiFI_data_dir, str(version) + "_proximity_score_adata_" + str(para) + ".h5ad")
    ImputeHiFI_adata = sc.read(store_path)

    chr_cell_type_center_value_df = first_seen_probe_data['chr_cell_type_center_value_df']
    chr_name_list = first_seen_probe_data['chr_name_list']
    cell_number = first_seen_probe_data['cell_number']
    all_cell_dna_list = first_seen_probe_data['all_cell_dna_list']
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']
    chrom_index_to_ids_dict = first_seen_probe_data['chrom_index_to_ids_dict']
    chrom_ids_to_index_dict = first_seen_probe_data['chrom_ids_to_index_dict']
    adata_obs_df = first_seen_probe_data['adata_obs_df']
    spaital_dist_suffix = first_seen_probe_data['spaital_dist_suffix']
    prxoimtiy_score_suffix = first_seen_probe_data['prxoimtiy_score_suffix']
    Linear_impute_all_cell_dna_list = first_seen_probe_data['Linear_all_cell_dna_list']

    infer_3D_type = r"_part_loci_pair_weight_"
    loci_pair_type = 'no_diff'

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + "_Impute_data.pkl")

    finall_optimize_all_cell_dna_list, diff_genes = ImputeHiFI.generate_impute_dna_list.get_impute_dna_list(
        chr_bin_length_dic, chr_name_list, ImputeHiFI_adata,
        all_cell_dna_list, Linear_impute_all_cell_dna_list,
        chr_cell_type_center_value_df, chrom_ids_to_index_dict,
        loci_pair_type, kernerl_number=None)

    final_ImputeHiFI_data = copy.deepcopy(first_seen_probe_data)
    final_ImputeHiFI_data['all_cell_dna_list'] = finall_optimize_all_cell_dna_list
    final_ImputeHiFI_data['ledien_diff_gene'] = diff_genes
    final_ImputeHiFI_data['cell_number']=len(finall_optimize_all_cell_dna_list)

    basic.store_variable_from_pikle_file(store_path, final_ImputeHiFI_data)

    ### fill all nan with linear
    first_seen_probe_data_path = os.path.join(first_seen_probe_data_dir, 'first_seen_probe_data.pkl')
    first_seen_probe_data = basic.load_variable_from_pikle_file(first_seen_probe_data_path)

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + "_Impute_data.pkl")
    final_ImputeHiFI_data = basic.load_variable_from_pikle_file(store_path)
    Linear_impute_all_cell_dna_list = first_seen_probe_data['Linear_all_cell_dna_list']
    final_ImputeHiFI_data['Linear_all_cell_dna_list']= Linear_impute_all_cell_dna_list
    finall_optimize_all_cell_dna_list=final_ImputeHiFI_data['all_cell_dna_list']
    finall_optimize_all_cell_dna_list=ImputeHiFI.generate_impute_dna_list.fill_all_nan_chr_with_linear(finall_optimize_all_cell_dna_list,Linear_impute_all_cell_dna_list)
    final_ImputeHiFI_data['all_cell_dna_list']=finall_optimize_all_cell_dna_list
    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + "_Impute_data.pkl")
    basic.store_variable_from_pikle_file(store_path, final_ImputeHiFI_data)

    ### generate 3D spatial distance

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + "_Impute_data.pkl")

    final_ImputeHiFI_data = basic.load_variable_from_pikle_file(store_path)

    chr_name_list = first_seen_probe_data['chr_name_list']
    cell_number = first_seen_probe_data['cell_number']
    all_cell_dna_list = first_seen_probe_data['all_cell_dna_list']
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']
    chrom_index_to_ids_dict = first_seen_probe_data['chrom_index_to_ids_dict']
    adata_obs_df = first_seen_probe_data['adata_obs_df']
    spaital_dist_suffix = first_seen_probe_data['spaital_dist_suffix']
    prxoimtiy_score_suffix = first_seen_probe_data['prxoimtiy_score_suffix']

    Impute_all_cell_dna_list = final_ImputeHiFI_data['all_cell_dna_list']
    cell_number = final_ImputeHiFI_data['cell_number']

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + f"_{spaital_dist_suffix}.h5ad")

    ImputeHiFI.generate_spatial_distance_h5ad.generate_spatial_h5ad(chr_name_list, cell_number, Impute_all_cell_dna_list,
                                                                    chr_bin_length_dic, chrom_index_to_ids_dict,
                                                                    adata_obs_df, store_path, )

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + f"_{spaital_dist_suffix}.h5ad")


    ### generate proximity score

    importlib.reload(ImputeHiFI)
    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + f"_{spaital_dist_suffix}.h5ad")
    Impute_3D_spad_adata = sc.read(store_path)

    chr_cell_type_center_value_df = first_seen_probe_data['chr_cell_type_center_value_df']
    chr_name_list = first_seen_probe_data['chr_name_list']
    cell_number = first_seen_probe_data['cell_number']
    all_cell_dna_list = first_seen_probe_data['all_cell_dna_list']
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']
    chrom_index_to_ids_dict = first_seen_probe_data['chrom_index_to_ids_dict']
    adata_obs_df = first_seen_probe_data['adata_obs_df']
    spaital_dist_suffix = first_seen_probe_data['spaital_dist_suffix']
    prxoimtiy_score_suffix = first_seen_probe_data['prxoimtiy_score_suffix']

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + f"_{prxoimtiy_score_suffix}.h5ad")
    ImputeHiFI.generate_proximity_score_h5ad.generate_proximity_h5ad(Impute_3D_spad_adata, chr_name_list,
                                                                     chr_cell_type_center_value_df,
                                                                     chr_bin_length_dic, adata_obs_df, store_path,
                                                                     kernerl_number=None)

    Impute_3D_proxi_adata = sc.read(store_path)

def run_ImputeHiFI_mode_3(first_seen_probe_data_dir,neighbor_cell_num=100):
    # data_dir = r"/mnt/disk1/scfan/data/Cai_21_Nature"
    # current_probe_data_dir = r"/mnt/disk1/scfan/data/Cai_21_Nature/DNA_25kb"
    # first_seen_probe_data_dir = r"/mnt/disk1/scfan/data/Cai_21_Nature/DNA_25kb/Impute/"
    basic.recursive_mkdir(first_seen_probe_data_dir)

    ImputeHiFI_data_dir = os.path.join(first_seen_probe_data_dir, 'ImputeHiFI')
    basic.recursive_mkdir(ImputeHiFI_data_dir)

    ### proxi adata

    FISH_RMSD_name = 'All_cell_RMSD_dict'
    version = 'no_HiC_RMSD_not_use_mean_count'

    first_seen_probe_data_path = os.path.join(first_seen_probe_data_dir, 'first_seen_probe_data.pkl')
    first_seen_probe_data = basic.load_variable_from_pikle_file(first_seen_probe_data_path)

    chr_cell_type_center_value_df = first_seen_probe_data['chr_cell_type_center_value_df']
    chr_name_list = first_seen_probe_data['chr_name_list']
    cell_name_list=first_seen_probe_data['cell_name_list']
    cell_number = first_seen_probe_data['cell_number']
    all_cell_dna_list = first_seen_probe_data['all_cell_dna_list']
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']
    chrom_index_to_ids_dict = first_seen_probe_data['chrom_index_to_ids_dict']
    adata_obs_df = first_seen_probe_data['adata_obs_df']
    spaital_dist_suffix = first_seen_probe_data['spaital_dist_suffix']
    prxoimtiy_score_suffix = first_seen_probe_data['prxoimtiy_score_suffix']
    missing_rate_df = first_seen_probe_data['missing_rate_df']
    FISH_nearest_neighbors_dic = first_seen_probe_data[FISH_RMSD_name]

    store_path = os.path.join(first_seen_probe_data_dir, f'first_seen_probe_data_{prxoimtiy_score_suffix}.h5ad')
    DNA_FISH_proxi_adata = sc.read(store_path)

    importlib.reload(ImputeHiFI_no_HiC)
    base_coef = 0.001
    HiC_weight = 0
    FISH_weight = 1
    short_range_HiC_weight_bound = 0.0
    long_range_HiC_weight_bound = 0.0
    all_range_diag_sum_weight_HiC_bound = 0.0
    short_range_diag = 0
    initial_guess = [HiC_weight, FISH_weight]
    verbose = False
    background_cell_num = neighbor_cell_num
    use_mean_count = 0
    para = ('bg_n' + str(background_cell_num) + '_sdn' + str(short_range_diag) + '_sd' + str(
        all_range_diag_sum_weight_HiC_bound) +
            '_sr' + str(short_range_HiC_weight_bound) + '_lr' + str(long_range_HiC_weight_bound))

    store_path = os.path.join(ImputeHiFI_data_dir, str(version) + "_proximity_score_adata_" + str(para) + ".h5ad")

    ImputeHiFI_no_HiC.ImputeHiFi_mode2_mode3.get_Impute_proximity_score(DNA_FISH_proxi_adata, missing_rate_df,
                                                                        cell_name_list, chr_bin_length_dic, base_coef,
                                                                        HiC_weight, FISH_weight, initial_guess,
                                                                        short_range_HiC_weight_bound,
                                                                        long_range_HiC_weight_bound,
                                                                        all_range_diag_sum_weight_HiC_bound,
                                                                        short_range_diag, background_cell_num,
                                                                        use_mean_count, FISH_nearest_neighbors_dic,
                                                                        store_path, kernerl_number=20)


    ### Impute 3D data

    first_seen_probe_data.keys()

    importlib.reload(ImputeHiFI)
    store_path = os.path.join(ImputeHiFI_data_dir, str(version) + "_proximity_score_adata_" + str(para) + ".h5ad")
    ImputeHiFI_adata = sc.read(store_path)

    chr_cell_type_center_value_df = first_seen_probe_data['chr_cell_type_center_value_df']
    chr_name_list = first_seen_probe_data['chr_name_list']
    cell_number = first_seen_probe_data['cell_number']
    all_cell_dna_list = first_seen_probe_data['all_cell_dna_list']
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']
    chrom_index_to_ids_dict = first_seen_probe_data['chrom_index_to_ids_dict']
    chrom_ids_to_index_dict = first_seen_probe_data['chrom_ids_to_index_dict']
    adata_obs_df = first_seen_probe_data['adata_obs_df']
    spaital_dist_suffix = first_seen_probe_data['spaital_dist_suffix']
    prxoimtiy_score_suffix = first_seen_probe_data['prxoimtiy_score_suffix']
    Linear_impute_all_cell_dna_list = first_seen_probe_data['Linear_all_cell_dna_list']

    infer_3D_type = r"_part_loci_pair_weight_"
    loci_pair_type = 'no_diff'

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + "_Impute_data.pkl")

    finall_optimize_all_cell_dna_list, diff_genes = ImputeHiFI.generate_impute_dna_list.get_impute_dna_list(
        chr_bin_length_dic, chr_name_list, ImputeHiFI_adata,
        all_cell_dna_list, Linear_impute_all_cell_dna_list,
        chr_cell_type_center_value_df, chrom_ids_to_index_dict,
        loci_pair_type, kernerl_number=None)

    final_ImputeHiFI_data = copy.deepcopy(first_seen_probe_data)
    final_ImputeHiFI_data['all_cell_dna_list'] = finall_optimize_all_cell_dna_list
    final_ImputeHiFI_data['ledien_diff_gene'] = diff_genes
    final_ImputeHiFI_data['cell_number']=len(finall_optimize_all_cell_dna_list)

    basic.store_variable_from_pikle_file(store_path, final_ImputeHiFI_data)

    ### fill all nan with linear
    first_seen_probe_data_path = os.path.join(first_seen_probe_data_dir, 'first_seen_probe_data.pkl')
    first_seen_probe_data = basic.load_variable_from_pikle_file(first_seen_probe_data_path)

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + "_Impute_data.pkl")
    final_ImputeHiFI_data = basic.load_variable_from_pikle_file(store_path)
    Linear_impute_all_cell_dna_list = first_seen_probe_data['Linear_all_cell_dna_list']
    final_ImputeHiFI_data['Linear_all_cell_dna_list']= Linear_impute_all_cell_dna_list
    finall_optimize_all_cell_dna_list=final_ImputeHiFI_data['all_cell_dna_list']
    finall_optimize_all_cell_dna_list=ImputeHiFI.generate_impute_dna_list.fill_all_nan_chr_with_linear(finall_optimize_all_cell_dna_list,Linear_impute_all_cell_dna_list)
    final_ImputeHiFI_data['all_cell_dna_list']=finall_optimize_all_cell_dna_list
    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + "_Impute_data.pkl")
    basic.store_variable_from_pikle_file(store_path, final_ImputeHiFI_data)

    ### generate 3D spatial distance

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + "_Impute_data.pkl")

    final_ImputeHiFI_data = basic.load_variable_from_pikle_file(store_path)

    chr_name_list = first_seen_probe_data['chr_name_list']
    cell_number = first_seen_probe_data['cell_number']
    all_cell_dna_list = first_seen_probe_data['all_cell_dna_list']
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']
    chrom_index_to_ids_dict = first_seen_probe_data['chrom_index_to_ids_dict']
    adata_obs_df = first_seen_probe_data['adata_obs_df']
    spaital_dist_suffix = first_seen_probe_data['spaital_dist_suffix']
    prxoimtiy_score_suffix = first_seen_probe_data['prxoimtiy_score_suffix']

    Impute_all_cell_dna_list = final_ImputeHiFI_data['all_cell_dna_list']
    cell_number = final_ImputeHiFI_data['cell_number']

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + f"_{spaital_dist_suffix}.h5ad")

    ImputeHiFI.generate_spatial_distance_h5ad.generate_spatial_h5ad(chr_name_list, cell_number, Impute_all_cell_dna_list,
                                                                    chr_bin_length_dic, chrom_index_to_ids_dict,
                                                                    adata_obs_df, store_path, )

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + f"_{spaital_dist_suffix}.h5ad")


    ### generate proximity score

    importlib.reload(ImputeHiFI)
    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + f"_{spaital_dist_suffix}.h5ad")
    Impute_3D_spad_adata = sc.read(store_path)

    chr_cell_type_center_value_df = first_seen_probe_data['chr_cell_type_center_value_df']
    chr_name_list = first_seen_probe_data['chr_name_list']
    cell_number = first_seen_probe_data['cell_number']
    all_cell_dna_list = first_seen_probe_data['all_cell_dna_list']
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']
    chrom_index_to_ids_dict = first_seen_probe_data['chrom_index_to_ids_dict']
    adata_obs_df = first_seen_probe_data['adata_obs_df']
    spaital_dist_suffix = first_seen_probe_data['spaital_dist_suffix']
    prxoimtiy_score_suffix = first_seen_probe_data['prxoimtiy_score_suffix']

    store_path = os.path.join(ImputeHiFI_data_dir,
                              version + str(para) + infer_3D_type + loci_pair_type + f"_{prxoimtiy_score_suffix}.h5ad")
    ImputeHiFI.generate_proximity_score_h5ad.generate_proximity_h5ad(Impute_3D_spad_adata, chr_name_list,
                                                                     chr_cell_type_center_value_df,
                                                                     chr_bin_length_dic, adata_obs_df, store_path,
                                                                     kernerl_number=None)

    Impute_3D_proxi_adata = sc.read(store_path)

def display_raw_data(first_seen_probe_data_dir,one_FISH_name,chr_name):
    prxoimtiy_score_suffix='proximity_score'
    store_path = os.path.join(first_seen_probe_data_dir, f'first_seen_probe_data_{prxoimtiy_score_suffix}.h5ad')
    DNA_FISH_proxi_adata = sc.read(store_path)
    # one_FISH_name = '0_3_1_rep1'
    # chr_name = 'chr1'
    first_seen_probe_data_path = os.path.join(first_seen_probe_data_dir, 'first_seen_probe_data.pkl')
    first_seen_probe_data = basic.load_variable_from_pikle_file(first_seen_probe_data_path)
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']

    one_array = ImputeHiFI.get_proximity_array(DNA_FISH_proxi_adata, one_FISH_name, chr_name, chr_bin_length_dic,
                                               set_min_limit=True, min_limit=0.001, set_max_limit=False, max_limit=1.0)
    ImputeHiFI.plot_proximity(one_array, title=one_FISH_name + ' ' + chr_name)

def display_imputed_data(first_seen_probe_data_dir,one_FISH_name,chr_name,file_name=None,version = 'no_HiC_RNA_FISH_RMSD_not_use_mean_count',):
    prxoimtiy_score_suffix='proximity_score'
    first_seen_probe_data_path = os.path.join(first_seen_probe_data_dir, 'first_seen_probe_data.pkl')
    first_seen_probe_data = basic.load_variable_from_pikle_file(first_seen_probe_data_path)
    chr_bin_length_dic = first_seen_probe_data['chr_bin_length_dic']

    HiC_weight = 0
    FISH_weight = 1
    short_range_HiC_weight_bound = 0.0
    long_range_HiC_weight_bound = 0.0
    all_range_diag_sum_weight_HiC_bound = 0.0
    short_range_diag = 0
    initial_guess = [HiC_weight, FISH_weight]
    verbose = False
    background_cell_num = 100
    use_mean_count = 0
    para = ('bg_n' + str(background_cell_num) + '_sdn' + str(short_range_diag) + '_sd' + str(
        all_range_diag_sum_weight_HiC_bound) +
            '_sr' + str(short_range_HiC_weight_bound) + '_lr' + str(long_range_HiC_weight_bound))
    infer_3D_type = r"_part_loci_pair_weight_"
    loci_pair_type = 'no_diff'
    if file_name is None:
        ImputeHiFI_data_dir = os.path.join(first_seen_probe_data_dir, 'ImputeHiFI')
        store_path = os.path.join(ImputeHiFI_data_dir,
                                  version + str(para) + infer_3D_type + loci_pair_type + f"_{prxoimtiy_score_suffix}.h5ad")
    else:
        store_path = os.path.join(first_seen_probe_data_dir, file_name)
    DNA_FISH_proxi_adata = sc.read(store_path)


    one_array = ImputeHiFI.get_proximity_array(DNA_FISH_proxi_adata, one_FISH_name, chr_name, chr_bin_length_dic,
                                               set_min_limit=True, min_limit=0.001, set_max_limit=False, max_limit=1.0)
    ImputeHiFI.plot_proximity(one_array, title=one_FISH_name + ' ' + chr_name)


import setuptools

if __name__ == "__main__":
    with open('requirements.txt') as f:
        required = f.read().splitlines()
    setuptools.setup(name="ImputeHiFI",version="0.1.3",install_requires=required,)



import os
import time
from subprocess import  call
import numpy as np
import pandas as pd
import seaborn as sns
from scipy import stats
import seaborn as sns
import matplotlib.pyplot as plt
import scipy as sc
from matplotlib.colors import ListedColormap
import scipy.sparse as sparse
import importlib
import copy
import gc
import datetime
from scipy.stats import zscore
import sys
import scipy.stats as st
import math
from scipy import optimize
from tqdm import tqdm
from scipy.spatial.distance import pdist, squareform
import scanpy as sc
import warnings
from scipy.interpolate import griddata

warnings.filterwarnings('ignore')
print('v6 similar FISH no HiC')
#we can use FISH_nearest_neighbors_names_dic to specify what are the nearest neighbors of each cell under each chromosome.
#According to observations, whether using mean_count or bg_all_count will have a greater impact on the clustering effect.

#In this version, this becomes optional through the use_mean_count parameter.


def infer_distance_matrix(matrix_array, sigma):
    """
    Infers the distance matrix from a given matrix computed using the Gaussian function.

    Parameters:
        matrix_array (np.array): The matrix computed by the Gaussian function.
        sigma (float): The standard deviation of the Gaussian function.

    Returns:
        dist_matrix (np.array): The inferred distance matrix.
    """
    # Ensure all values in matrix_array are positive
    if np.any(matrix_array <= 0):
        raise ValueError("All elements in matrix_array must be positive.")

    # Compute the square of the distances
    d_squared = -np.log2(matrix_array) * sigma**2

    # Compute the distance matrix by taking the square root of the squared distances
    dist_matrix = np.sqrt(d_squared)

    return dist_matrix




def convert_array_to_bin_pair_df(chr_name,bin_length,one_cell_chr_array,data=0):
    matrix = np.ones((bin_length, bin_length))
    matrix = np.triu(matrix, k=1)

    from scipy.sparse import coo_matrix
    matrix_sparse = coo_matrix(matrix)

    df = pd.DataFrame({'col': matrix_sparse.col, 'row': matrix_sparse.row, 'data': matrix_sparse.data})

    df['col'] = df['col'].astype(str)
    df['row'] = df['row'].astype(str)
    df['data']=data
    df['chr_bin_bin']=chr_name+'_'+df['row']+'_'+df['col']
    Example_df=df.copy()
    Example_df.index=Example_df['chr_bin_bin']

    one_cell_chr_array=np.triu(one_cell_chr_array, k=1)
    one_cell_chr_array_sparse = coo_matrix(one_cell_chr_array)

    df = pd.DataFrame({'col': one_cell_chr_array_sparse.col, 'row': one_cell_chr_array_sparse.row, 'data': one_cell_chr_array_sparse.data})
    df['col'] = df['col'].astype(str)
    df['row'] = df['row'].astype(str)
    df['chr_bin_bin']=chr_name+'_'+df['row']+'_'+df['col']
    one_cell_chr_df=df.copy()
    one_cell_chr_df.index=one_cell_chr_df['chr_bin_bin']
    Example_df.loc[one_cell_chr_df.index,'data']=one_cell_chr_df['data'].values
    return Example_df
def generate_Example_bin_pair_df(chr_name,bin_length,data=0):
    matrix = np.ones((bin_length, bin_length))
    matrix = np.triu(matrix, k=1)

    from scipy.sparse import coo_matrix
    matrix_sparse = coo_matrix(matrix)

    df = pd.DataFrame({'col': matrix_sparse.col, 'row': matrix_sparse.row, 'data': matrix_sparse.data})
    df['data']=data
    df['col'] = df['col'].astype(str)
    df['row'] = df['row'].astype(str)
    df['chr_bin_bin']=chr_name+'_'+df['row']+'_'+df['col']
    Example_df=df.copy()
    Example_df.index=Example_df['chr_bin_bin']
    return Example_df

def convert_bin_pair_df_to_array(bin_length, Example_df):
    # 将'chr_bin_bin'列分解为'chr_name', 'row', 'col'
    Example_df[['chr_name', 'row', 'col']] = Example_df['chr_bin_bin'].str.split('_', expand=True)

    # 将'row'和'col'列的数据类型转换为整数，以便用于创建coo_matrix
    Example_df['row'] = Example_df['row'].astype(int)
    Example_df['col'] = Example_df['col'].astype(int)

    # 使用'row'，'col'，和'data'列创建coo_matrix
    from scipy.sparse import coo_matrix
    matrix_sparse = coo_matrix((Example_df['data'], (Example_df['row'], Example_df['col'])), shape=(bin_length, bin_length))

    # 将coo_matrix转换为数组
    array = matrix_sparse.toarray()

    return array

def generate_DNA_FISH_df(one_FISH_name,chr_name,missing_rate_df,DNA_FISH_adata):
    one_cell_chr_info_df = missing_rate_df.loc[[(one_FISH_name, chr_name)]]
    bin_length = one_cell_chr_info_df['total_probe_number'].values[0]
    Example_df = generate_Example_bin_pair_df(chr_name, bin_length)
    Common_Example_df = copy.deepcopy(Example_df)
    Common_Example_df['diag'] = Common_Example_df['col'].astype(int) - Common_Example_df['row'].astype(int)
    diag_array = np.unique(Common_Example_df['diag'].values)

    Example_df['data'] = DNA_FISH_adata[one_FISH_name, Example_df.index].copy().X[0]
    all_count = np.nansum(Example_df['data'].values)
    Example_df['prop'] = Example_df['data'] / all_count
    # norm diag prop
    # one_diag=1
    diag_sum_prop = {}
    for one_diag in diag_array:
        one_diag_index = Common_Example_df[Common_Example_df['diag'] == one_diag].index
        temp_sum = np.nansum(Example_df.loc[one_diag_index]['prop'].values)
        diag_sum_prop[one_diag] = temp_sum
        if (temp_sum == 0):
            Example_df.loc[one_diag_index, 'diag_prop'] = 0.0
        else:
            Example_df.loc[one_diag_index, 'diag_prop'] = Example_df.loc[one_diag_index]['prop'] / temp_sum

    DNA_FISH_df = Example_df.copy()
    return DNA_FISH_df,diag_sum_prop,all_count,Common_Example_df,diag_array,bin_length
def generate_HiC_df(diag_array,Common_Example_df,multi_HiC_name_list,chr_name,bin_length,weight_multi_HiC_RNAcoef_array,HiC_adata):
    temp_HiC_Example_df_dict = {}
    # one_HiC_ids=0
    # one_HiC_name=multi_HiC_name_list[one_HiC_ids]
    for one_HiC_ids, one_HiC_name in enumerate(multi_HiC_name_list):
        Example_df = generate_Example_bin_pair_df(chr_name, bin_length)

        one_HiC_weight = weight_multi_HiC_RNAcoef_array[one_HiC_ids]
        Example_df['data'] = HiC_adata[one_HiC_name, Example_df.index].copy().X[0]
        all_count = np.nansum(Example_df['data'].values)
        Example_df['prop'] = Example_df['data'] / all_count
        Example_df['weight_prop'] = Example_df['prop'] * one_HiC_weight
        temp_HiC_Example_df_dict[one_HiC_name] = Example_df
    Example_df = generate_Example_bin_pair_df(chr_name, bin_length)
    Example_df['prop'] = 0.0
    for one_HiC_name, one_HiC_df in temp_HiC_Example_df_dict.items():
        Example_df['prop'] = Example_df['prop'] + one_HiC_df['weight_prop']

    # norm diag prop
    # one_diag=1
    diag_sum_prop = {}
    for one_diag in diag_array:
        one_diag_index = Common_Example_df[Common_Example_df['diag'] == one_diag].index
        temp_sum = np.nansum(Example_df.loc[one_diag_index]['prop'].values)
        diag_sum_prop[one_diag] = temp_sum
        if (temp_sum == 0):
            Example_df.loc[one_diag_index, 'diag_prop'] = 0.0
        else:
            Example_df.loc[one_diag_index, 'diag_prop'] = Example_df.loc[one_diag_index]['prop'] / temp_sum

    HiC_df = copy.deepcopy(Example_df)
    HiC_diag_sum_prop_dict = copy.deepcopy(diag_sum_prop)
    return HiC_df, HiC_diag_sum_prop_dict
def generate_background_df(diag_array,Common_Example_df,one_FISH_name,chr_name,bin_length,DNA_FISH_adata,FISH_nearest_neighbors_names_dic,background_cell_num):
    # background_cell_num=50
    Example_df = generate_Example_bin_pair_df(chr_name, bin_length)
    temp_array = DNA_FISH_adata[
        FISH_nearest_neighbors_names_dic[chr_name][one_FISH_name][0:background_cell_num], Example_df.index].copy().X

    Example_df['data'] = np.nanmedian(temp_array, axis=0)

    all_count = np.nansum(Example_df['data'].values)

    Example_df['prop'] = Example_df['data'] / all_count

    # norm diag prop
    # one_diag=1
    diag_sum_prop = {}
    for one_diag in diag_array:
        one_diag_index = Common_Example_df[Common_Example_df['diag'] == one_diag].index
        temp_sum = np.nansum(Example_df.loc[one_diag_index]['prop'].values)
        diag_sum_prop[one_diag] = temp_sum
        if (temp_sum == 0):
            Example_df.loc[one_diag_index, 'diag_prop'] = 0.0
        else:
            Example_df.loc[one_diag_index, 'diag_prop'] = Example_df.loc[one_diag_index]['prop'] / temp_sum

    bg_df = copy.deepcopy(Example_df)
    bg_diag_sum_prop_dict = copy.deepcopy(diag_sum_prop)
    return bg_df, bg_diag_sum_prop_dict,all_count
def generate_Impute_DNA_FISH_df(DNA_FISH_df,HiC_df,bg_df,Common_Example_df,diag_array,HiC_diag_sum_prop_dict,bg_diag_sum_prop_dict,DNA_FISH_weight,HiC_weight,background_weight,HiC_diag_weight,background_diag_weight):
    DNA_FISH_df['prop'].fillna(0, inplace=True)
    DNA_FISH_df['diag_prop'].fillna(0, inplace=True)

    HiC_df['prop'].fillna(0, inplace=True)
    HiC_df['diag_prop'].fillna(0, inplace=True)

    bg_df['prop'].fillna(0, inplace=True)
    bg_df['diag_prop'].fillna(0, inplace=True)

    merged_diag_sum_prop_dict = {}
    for one_diag in diag_array:
        temp_sum_prop = HiC_diag_sum_prop_dict[one_diag] * HiC_diag_weight + bg_diag_sum_prop_dict[
            one_diag] * background_diag_weight
        merged_diag_sum_prop_dict[one_diag] = temp_sum_prop

    Impute_DNA_FISH_df = copy.deepcopy(DNA_FISH_df)

    Impute_DNA_FISH_df['Impute_diag_prop'] = 0
    Impute_DNA_FISH_df['Impute_prop'] = 0

    for one_diag in diag_array:
        one_diag_index = Common_Example_df[Common_Example_df['diag'] == one_diag].index
        temp_DNA_FISH_diag = (DNA_FISH_df.loc[one_diag_index, 'diag_prop']) * DNA_FISH_weight
        temp_HiC_diag = (HiC_df.loc[one_diag_index, 'diag_prop']) * HiC_weight
        temp_bg_diag = (bg_df.loc[one_diag_index, 'diag_prop']) * background_weight

        Impute_DNA_FISH_df.loc[one_diag_index, 'Impute_diag_prop'] = temp_DNA_FISH_diag + temp_HiC_diag + temp_bg_diag
        #     print(one_diag,np.sum(Impute_DNA_FISH_df.loc[one_diag_index,'Impute_diag_prop']))
        merged_one_diag_sum_prop = merged_diag_sum_prop_dict[one_diag]
        Impute_DNA_FISH_df.loc[one_diag_index, 'Impute_prop'] = Impute_DNA_FISH_df.loc[
                                                                    one_diag_index, 'Impute_diag_prop'] * merged_one_diag_sum_prop

    prop_count = np.sum(Impute_DNA_FISH_df['Impute_prop'])
    Impute_DNA_FISH_df['Impute_prop_norm'] = Impute_DNA_FISH_df['Impute_prop'] / prop_count
    return Impute_DNA_FISH_df
def generate_sample_df(chr_name,bin_length,sample_range,sample_time,sample_prop):
    sample_result = np.random.choice(sample_range, size=sample_range * sample_time, replace=True, p=sample_prop)

    Example_df = generate_Example_bin_pair_df(chr_name, bin_length, data=1)
    temp_Example_df = Example_df.iloc[sample_result]
    temp_Example_df.reset_index(drop=True, inplace=True)

    sum_by_chr_bin_bin = temp_Example_df.groupby('chr_bin_bin')['data'].sum().reset_index()

    sum_by_chr_bin_bin.index = sum_by_chr_bin_bin['chr_bin_bin']

    Example_df.loc[sum_by_chr_bin_bin.index, 'data'] = sum_by_chr_bin_bin['data']
    return Example_df

def fill_nan_with_interpolation(distmap_list_median):

    nan_locs = np.isnan(distmap_list_median)
    not_nan_locs = ~nan_locs
    not_nan_values = distmap_list_median[not_nan_locs]
    not_nan_positions = np.array(np.nonzero(not_nan_locs)).T


    nan_positions = np.array(np.nonzero(nan_locs)).T


    interpolated_values = griddata(not_nan_positions, not_nan_values, nan_positions, method='nearest')


    filled_distmap_list_median = distmap_list_median.copy()
    filled_distmap_list_median[nan_locs] = interpolated_values

    return filled_distmap_list_median

import numpy as np
from scipy.optimize import minimize
from scipy.optimize import minimize

# 定义目标函数
def objective(weights, P_FISH_observed, P_HiC_observed, P_neighbor_FISH_observed):
    a, b = weights
    a=0
    b=weights[1]
    prediction = a * P_HiC_observed + b * P_neighbor_FISH_observed
    return np.sum((prediction - P_FISH_observed) ** 2)
def get_predicted(weights, P_FISH_observed, P_HiC_observed, P_neighbor_FISH_observed):
    a, b = weights
    a=0
    b=weights[1]
    prediction = a * P_HiC_observed + b * P_neighbor_FISH_observed
    return prediction
# 使用最佳权重填补缺失数据
def impute_missing_data(weights, P_HiC_missing, P_neighbor_FISH_missing):
    a, b = weights
    a=0
    b=weights[1]
    return a * P_HiC_missing + b * P_neighbor_FISH_missing
def safe_normalize(array):
    array_sum = array.sum()
    if array_sum == 0:
        return np.zeros_like(array)
    else:
        return array / array_sum
def extract_data(one_diag_index, DNA_FISH_df, HiC_df, bg_df):
    P_FISH = DNA_FISH_df.loc[one_diag_index, 'diag_prop'].values
    P_HiC = HiC_df.loc[one_diag_index, 'diag_prop'].values
    P_neighbor_FISH = bg_df.loc[one_diag_index, 'diag_prop'].values
    P_neighbor_FISH = np.nan_to_num(P_neighbor_FISH, 0)
    # 区分观测到的数据和缺失的数据
    observed_mask = ~DNA_FISH_df.loc[one_diag_index, 'data'].isna()
    missing_mask = DNA_FISH_df.loc[one_diag_index, 'data'].isna()

    # 创建观测到的数据和缺失的数据的数组
    P_FISH_observed = DNA_FISH_df.loc[one_diag_index[observed_mask], 'diag_prop'].values
    # P_FISH_missing = DNA_FISH_df.loc[one_diag_index[missing_mask], 'diag_prop'].values

    # 如果需要，也可以创建包含其他列的DataFrame
    # P_FISH_observed_df = DNA_FISH_df.loc[one_diag_index[observed_mask]]
    # P_FISH_missing_df = DNA_FISH_df.loc[one_diag_index[missing_mask]]

    # 现在你有了观测到的数据和缺失的数据的索引，可以用于后续的插补步骤
    # 提取 P_HiC 和 P_neighbor_FISH 中观测到的数据和缺失的数据
    P_HiC_observed = HiC_df.loc[one_diag_index[observed_mask], 'diag_prop'].values
    P_HiC_missing = HiC_df.loc[one_diag_index[missing_mask], 'diag_prop'].values

    P_neighbor_FISH_observed = bg_df.loc[one_diag_index[observed_mask], 'diag_prop'].values
    P_neighbor_FISH_missing = bg_df.loc[one_diag_index[missing_mask], 'diag_prop'].values
    P_neighbor_FISH_observed = np.nan_to_num(P_neighbor_FISH_observed, 0)
    P_neighbor_FISH_missing = np.nan_to_num(P_neighbor_FISH_missing, 0)

    # P_FISH_observed_norm = safe_normalize(P_FISH_observed)
    # P_HiC_observed_norm = safe_normalize(P_HiC_observed)
    # P_HiC_missing_norm = safe_normalize(P_HiC_missing)
    # P_neighbor_FISH_observed_norm = safe_normalize(P_neighbor_FISH_observed)
    # P_neighbor_FISH_missing_norm = safe_normalize(P_neighbor_FISH_missing)
    #
    # # 如果需要，也可以创建包含其他列的DataFrame
    # P_HiC_observed_df = HiC_df.loc[one_diag_index[observed_mask]]
    # P_HiC_missing_df = HiC_df.loc[one_diag_index[missing_mask]]
    #
    # P_neighbor_FISH_observed_df = bg_df.loc[one_diag_index[observed_mask]]
    # P_neighbor_FISH_missing_df = bg_df.loc[one_diag_index[missing_mask]]
    return P_FISH_observed, P_HiC_observed, P_HiC_missing,P_neighbor_FISH_observed, P_neighbor_FISH_missing,observed_mask, missing_mask

# 执行优化
def perform_optimization(objective, initial_guess, P_FISH_observed_norm, P_HiC_observed_norm,
                P_neighbor_FISH_observed_norm, constraints, diag_sum_constraints):
    result = minimize(
        objective,
        initial_guess,
        args=(P_FISH_observed_norm, P_HiC_observed_norm, P_neighbor_FISH_observed_norm),
        constraints=constraints,
        method='SLSQP'
    )

    fitted_weights = result.x
    result = minimize(
        objective,
        initial_guess,
        args=(P_FISH_observed_norm, P_HiC_observed_norm, P_neighbor_FISH_observed_norm),
        constraints=diag_sum_constraints,
        method='SLSQP'
    )
    diag_fitted_weights = result.x
    return fitted_weights, diag_fitted_weights


def optimize_and_impute_data(
    diag_array, Common_Example_df, DNA_FISH_df, HiC_df, bg_df,
    initial_guess, short_range_diag, short_range_HiC_weight_bound,
    long_range_HiC_weight_bound, all_range_diag_sum_weight_HiC_bound,
    HiC_diag_sum_prop_dict, bg_diag_sum_prop_dict, bg_all_count,use_mean_count,
):

    # Set up the constraints for the optimization problem
    diag_sum_constraints = (
        {'type': 'eq', 'fun': lambda weights: weights[0]},  # a =0 no Hi-C weight
        {'type': 'eq', 'fun': lambda weights: weights[1]-1},  # b = 1
        # {'type': 'ineq', 'fun': lambda weights: weights[1]},  # b >= 0
        # {'type': 'ineq', 'fun': lambda weights: 1 - weights[1]},  # b <= 1
        # {'type': 'eq', 'fun': lambda weights: weights[0] + weights[1] - 1}  # a + b = 1
    )

    # Initialize dictionaries to store the weights, losses, and updated data
    diag_learning_weight_dict = {}
    diag_learning_weight_for_diag_sum_dict = {}
    diag_init_loss = {}
    diag_learning_loss = {}
    DNA_Impute_FISH_df = DNA_FISH_df.copy()
    DNA_Impute_diag_sum_prop_dict = {}

    # Iterate over each diagonal in the diag_array
    for one_diag in diag_array:
        # Set the constraints based on the diagonal range
        if one_diag <= short_range_diag:
            min_HiC_weight = short_range_HiC_weight_bound
        else:
            min_HiC_weight = long_range_HiC_weight_bound
        constraints = (
            {'type': 'eq', 'fun': lambda weights: weights[0] },  # no Hi-C weight
            {'type': 'ineq', 'fun': lambda weights: weights[1]},  # b >= 0
            # {'type': 'ineq', 'fun': lambda weights: 1-weights[1]},  # b <= 1
            # {'type': 'eq', 'fun': lambda weights: weights[0] + weights[1] - 1}  # a + b = 1
        )

        # Get the indices for the current diagonal
        one_diag_index = Common_Example_df[Common_Example_df['diag'] == one_diag].index

        # Extract the observed and missing data for FISH, HiC, and neighboring FISH
        P_FISH_observed, P_HiC_observed, P_HiC_missing,P_neighbor_FISH_observed, P_neighbor_FISH_missing,observed_mask, missing_mask = extract_data(
            one_diag_index, DNA_FISH_df, HiC_df, bg_df
        )

        # Normalize the observed and missing data
        P_FISH_observed_norm = safe_normalize(P_FISH_observed)
        P_HiC_observed_norm = safe_normalize(P_HiC_observed)
        P_HiC_missing_norm = safe_normalize(P_HiC_missing)
        P_neighbor_FISH_observed_norm = safe_normalize(P_neighbor_FISH_observed)
        P_neighbor_FISH_missing_norm = safe_normalize(P_neighbor_FISH_missing)

        # Optimize the weights if there are observed data points
        if len(P_FISH_observed) == 0 or len(one_diag_index) == 1 or P_FISH_observed.sum() == 0:
            fitted_weights = initial_guess
            diag_fitted_weights = initial_guess
        else:
            fitted_weights, diag_fitted_weights = perform_optimization(
                objective, initial_guess, P_FISH_observed_norm, P_HiC_observed_norm,
                P_neighbor_FISH_observed_norm, constraints, diag_sum_constraints
            )

        # Impute the missing data using the optimized weights
        P_FISH_missing_filled = impute_missing_data(fitted_weights, P_HiC_missing_norm, P_neighbor_FISH_missing_norm)

        # Calculate the initial and learned losses for diagnostics
        diag_learning_weight_dict[one_diag] = fitted_weights
        diag_learning_weight_for_diag_sum_dict[one_diag] = diag_fitted_weights
        diag_init_loss[one_diag] = objective(initial_guess, P_FISH_observed_norm, P_HiC_observed_norm, P_neighbor_FISH_observed_norm)
        diag_learning_loss[one_diag] = objective(fitted_weights, P_FISH_observed_norm, P_HiC_observed_norm, P_neighbor_FISH_observed_norm)

        # Update the original data frame with the imputed data
        a, b =fitted_weights
        P_observed_impute_sum=a*P_HiC_observed.sum()+b*P_neighbor_FISH_observed.sum()
        if P_FISH_observed.sum()>0 and P_observed_impute_sum==0:
            observed_len=one_diag_index[observed_mask].shape[0]
            missing_len=one_diag_index[missing_mask].shape[0]
            P_observed_impute_sum=observed_len/(observed_len+missing_len)
            P_missing_impute_sum=1-P_observed_impute_sum
        else:
            P_missing_impute_sum=1-P_observed_impute_sum
        P_FISH_missing_filled=P_FISH_missing_filled*P_missing_impute_sum

        P_FISH_observed_filled=P_FISH_observed_norm*P_observed_impute_sum

        # 更新原始数据框中缺失数据的估计值

        DNA_Impute_FISH_df.loc[one_diag_index[missing_mask], 'diag_prop'] = P_FISH_missing_filled
        DNA_Impute_FISH_df.loc[one_diag_index[observed_mask], 'diag_prop'] = P_FISH_observed_filled
        temp_sum=DNA_Impute_FISH_df.loc[one_diag_index]['diag_prop'].sum()
        if temp_sum>0:
            DNA_Impute_FISH_df.loc[one_diag_index, 'diag_prop']=DNA_Impute_FISH_df.loc[one_diag_index, 'diag_prop']/temp_sum
        diag_sum_a,diag_sum_b=diag_fitted_weights
    #     DNA_Impute_diag_sum_prop_dict[one_diag]=HiC_diag_sum_prop_dict[one_diag]*a+bg_diag_sum_prop_dict[one_diag]*b
        DNA_Impute_diag_sum_prop_dict[one_diag]=HiC_diag_sum_prop_dict[one_diag]*diag_sum_a+bg_diag_sum_prop_dict[one_diag]*diag_sum_b

    DNA_Impute_diag_sum_prop_norm_dict = {}

    DNA_Impute_diag_sum_prop_array = np.array(
        [DNA_Impute_diag_sum_prop_dict[key] for key in DNA_Impute_diag_sum_prop_dict.keys()])

    temp_sum = DNA_Impute_diag_sum_prop_array.sum()
    for key, values in DNA_Impute_diag_sum_prop_dict.items():
        DNA_Impute_diag_sum_prop_norm_dict[key] = values / temp_sum

    DNA_Impute_FISH_df['Impute_data'] = DNA_Impute_FISH_df['data']
    for one_diag in diag_array:
        #     one_diag=1
        one_diag_index = Common_Example_df[Common_Example_df['diag'] == one_diag].index
        DNA_Impute_FISH_df.loc[one_diag_index, 'prop'] = DNA_Impute_diag_sum_prop_norm_dict[one_diag] * \
                                                         DNA_Impute_FISH_df.loc[one_diag_index, 'diag_prop']
        observed_mask = ~DNA_FISH_df.loc[one_diag_index, 'data'].isna()
        # missing_mask = DNA_FISH_df.loc[one_diag_index, 'data'].isna()
        temp_DNA_Impute_FISH_df = DNA_Impute_FISH_df.loc[one_diag_index[observed_mask]]
        count_series = temp_DNA_Impute_FISH_df['data'] / temp_DNA_Impute_FISH_df['prop']
        mean_count = count_series.mean()
        # diag_series_dict[one_diag] = mean_count

        if use_mean_count:
            if (np.isnan(mean_count)):
                DNA_Impute_FISH_df.loc[one_diag_index, 'Impute_data'] = DNA_Impute_FISH_df.loc[
                                                                            one_diag_index, 'prop'] * bg_all_count
            #         print(one_diag)
            else:
                DNA_Impute_FISH_df.loc[one_diag_index, 'Impute_data'] = DNA_Impute_FISH_df.loc[
                                                                               one_diag_index, 'prop'] * mean_count
        else:
            DNA_Impute_FISH_df.loc[one_diag_index, 'Impute_data'] = DNA_Impute_FISH_df.loc[one_diag_index, 'prop'] * bg_all_count

    DNA_Impute_FISH_df['data'] = DNA_Impute_FISH_df['Impute_data']
    # DNA_Impute_FISH_df['data'] = DNA_Impute_FISH_df['prop'] * bg_all_count
    # Impute_df_dict[chr_name] = DNA_Impute_FISH_df

    return DNA_Impute_FISH_df, diag_learning_weight_dict, diag_learning_weight_for_diag_sum_dict, diag_init_loss, diag_learning_loss
class ImputeHiFi_mode2_mode3(object):
    @staticmethod
    def generate_Impute_proximity_score_result_part(chr_bin_length_dic, initial_guess,short_range_HiC_weight_bound,
                               long_range_HiC_weight_bound, all_range_diag_sum_weight_HiC_bound,
                               short_range_diag, background_cell_num,
                               DNA_FISH_name_list, missing_rate_df, DNA_FISH_adata,
                               FISH_nearest_neighbors_names_dic, use_mean_count):
        Impute_DNA_FISH_df_dict = {}
        for one_FISH_name in tqdm(DNA_FISH_name_list):
            # chr_name='chr1'
            Impute_df_dict = {}
            for chr_name in chr_bin_length_dic.keys():
                DNA_FISH_df, diag_sum_prop, DNA_FISH_all_count, Common_Example_df, diag_array, bin_length = generate_DNA_FISH_df(
                    one_FISH_name, chr_name, missing_rate_df, DNA_FISH_adata)

                # HiC_df, HiC_diag_sum_prop_dict = ImputeHiFI.generate_HiC_df(diag_array, Common_Example_df,
                #                                                             multi_HiC_name_list, chr_name, bin_length,
                #                                                             weight_multi_HiC_RNAcoef_array, HiC_adata)

                bg_df, bg_diag_sum_prop_dict, bg_all_count = generate_background_df(diag_array,
                                                                                               Common_Example_df,
                                                                                               one_FISH_name, chr_name,
                                                                                               bin_length,
                                                                                               DNA_FISH_adata,
                                                                                               FISH_nearest_neighbors_names_dic,
                                                                                               background_cell_num)
                HiC_df = copy.deepcopy(bg_df)
                HiC_diag_sum_prop_dict = copy.deepcopy(bg_diag_sum_prop_dict)

                DNA_Impute_FISH_df, diag_learning_weight_dict, diag_learning_weight_for_diag_sum_dict, diag_init_loss, diag_learning_loss = optimize_and_impute_data(
                    diag_array, Common_Example_df, DNA_FISH_df, HiC_df, bg_df,
                    initial_guess, short_range_diag, short_range_HiC_weight_bound,
                    long_range_HiC_weight_bound, all_range_diag_sum_weight_HiC_bound,
                    HiC_diag_sum_prop_dict, bg_diag_sum_prop_dict, bg_all_count, use_mean_count)
                Impute_df_dict[chr_name] = DNA_Impute_FISH_df

            #         one_array=ImputeHiFI.convert_bin_pair_df_to_array(bin_length,DNA_FISH_df)
            #         cmap = plt.cm.seismic
            #         cmap.set_bad('#8a898a')
            #         sns.heatmap(one_array,cmap=cmap)
            #         plt.title(chr_name)
            #         plt.show()

            #         one_array=ImputeHiFI.convert_bin_pair_df_to_array(bin_length,DNA_Impute_FISH_df)
            #         cmap = plt.cm.seismic
            #         cmap.set_bad('#8a898a')
            #         sns.heatmap(one_array,cmap=cmap,vmax=1)
            #         plt.title(chr_name)
            #         plt.show()
            #         break
            one_cell_Impute_df = pd.concat(Impute_df_dict.values(), axis=0)
            Impute_DNA_FISH_df_dict[one_FISH_name] = one_cell_Impute_df
        return Impute_DNA_FISH_df_dict

    @staticmethod
    def get_Impute_proximity_score(DNA_FISH_adata,missing_rate_df,cell_name_list,chr_bin_length_dic,base_coef,HiC_weight,FISH_weight,initial_guess,short_range_HiC_weight_bound,
                                 long_range_HiC_weight_bound, all_range_diag_sum_weight_HiC_bound,
                                    short_range_diag, background_cell_num,use_mean_count,FISH_nearest_neighbors_dic,store_path,kernerl_number=32):
        FISH_nearest_neighbors_names_dic = FISH_nearest_neighbors_dic['nearest_neighbors_names_dic']

        parameter_dic = {}
        parameter_dic['base_coef'] = base_coef
        parameter_dic['HiC_weight'] = HiC_weight
        parameter_dic['FISH_weight'] = FISH_weight
        parameter_dic['short_range_HiC_weight_bound'] = short_range_HiC_weight_bound
        parameter_dic['long_range_HiC_weight_bound'] = long_range_HiC_weight_bound
        parameter_dic['all_range_diag_sum_weight_HiC_bound'] = all_range_diag_sum_weight_HiC_bound
        parameter_dic['short_range_diag'] = short_range_diag
        parameter_dic['background_cell_num'] = background_cell_num
        parameter_dic['use_mean_count'] = use_mean_count

        DNA_FISH_name_split_list = np.array_split(cell_name_list, kernerl_number)

        _domain_args = [(chr_bin_length_dic, initial_guess,short_range_HiC_weight_bound,
                               long_range_HiC_weight_bound, all_range_diag_sum_weight_HiC_bound,
                               short_range_diag, background_cell_num,
                               DNA_FISH_name_list, missing_rate_df, DNA_FISH_adata,
                               FISH_nearest_neighbors_names_dic, use_mean_count)
                        for DNA_FISH_name_list in DNA_FISH_name_split_list]
        import multiprocessing as mp

        with mp.Pool(kernerl_number) as domain_pool:
            domain_results = domain_pool.starmap(ImputeHiFi_mode2_mode3.generate_Impute_proximity_score_result_part, _domain_args)
            domain_pool.close()
            domain_pool.join()
            domain_pool.terminate()

        result_dic = {}
        for result in domain_results:
            print("Collect result")
            for cell_name, df in result.items():
                result_dic[cell_name] = df
        Example_dict = {}
        for chr_name, bin_length in chr_bin_length_dic.items():
            Example_dict[chr_name] = generate_Example_bin_pair_df(chr_name, bin_length)

        all_index_df = pd.concat(Example_dict.values(), axis=0)
        all_cell_df = pd.DataFrame(index=all_index_df.index)
        for cell_name, df in result_dic.items():
            all_cell_df[cell_name] = df.loc[all_cell_df.index, 'data'].values
        Impute_adata = sc.AnnData(all_cell_df.T)
        Impute_adata.obs = DNA_FISH_adata.obs.loc[Impute_adata.obs_names]
        Impute_adata.uns['parameter'] = parameter_dic

        Impute_adata.write(store_path)
def get_RNA_FISH_nearset_neighbors(data_dir):
    from scipy.spatial import cKDTree
    temp_path = os.path.join(data_dir, 'RNA', 'Cai_RNA_adata.h5ad')
    Cai_adata = sc.read(temp_path)
    adata = Cai_adata.copy()

    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.scale(adata)
    # sc.pp.highly_variable_genes(adata, n_top_genes=15000,min_mean=0.0125, max_mean=3, min_disp=0.5)
    # adata = adata[:, adata.var['highly_variable']]

    sc.pp.pca(adata, n_comps=30)
    # sc.pl.pca(adata,color='cell_type')
    sc.pp.neighbors(adata, n_pcs=30, n_neighbors=100, use_rep='X_pca')
    sc.tl.umap(adata)
    adata_tree = cKDTree(adata.obsm['X_umap'])

    # 对于 Transcriptom 数据集中的每个细胞，找到在 3D Genome 数据集中的最近邻

    neighbor_num = adata.shape[0]
    nearest_neighbors_distances, nearest_neighbors_indices = adata_tree.query(adata.obsm['X_umap'], k=neighbor_num)

    RNA_FISH_adata = adata.copy()
    RNA_FISH_nearest_neighbors_distances = nearest_neighbors_distances
    RNA_FISH_nearest_neighbors_indices = nearest_neighbors_indices
    RNA_FISH_nearest_neighbors_indices_dic = {}
    RNA_FISH_nearest_neighbors_names_dic = {}
    for idx, one_name in enumerate(adata.obs_names):
        RNA_FISH_nearest_neighbors_indices_dic[one_name] = RNA_FISH_nearest_neighbors_indices[idx][1:]
        RNA_FISH_nearest_neighbors_names_dic[one_name] = adata.obs_names[RNA_FISH_nearest_neighbors_indices[idx][1:]]
    RNA_FISH_nearest_neighbors_distances_dic = {}
    for idx, one_name in enumerate(adata.obs_names):
        RNA_FISH_nearest_neighbors_distances_dic[one_name] = RNA_FISH_nearest_neighbors_distances[idx][1:]
    return RNA_FISH_nearest_neighbors_names_dic